"""Module representing a collection of jobs."""
import os
import re
from typing import Dict, List, Optional, Tuple

from .job import Job
from .output_renderer import OutputRenderer


class JobCollection:
    """A group of jobs."""

    def __init__(self) -> None:
        """Create a new job collection with default options."""
        self.columns = [
            "JobIDRaw",
            "JobID",
            "State",
            "AllocCPUS",
            "TotalCPU",
            "Elapsed",
            "Timelimit",
            "REQMEM",
            "MaxRSS",
            "NNodes",
            "NTasks",
            "Partition",
        ]

        self.job_file_regex = re.compile(
            r"^.*?[_-](?P<jobid>(?P<job>[0-9]+)(_[0-9]+)?)(.out)?$"
        )
        self.job_regex = re.compile(r"^(?P<jobid>(?P<job>[0-9]+)(_[][\-0-9]+)?)$")

        self.jobs: Dict[str, Job] = {}
        self.renderer: Optional[OutputRenderer] = None
        self.dir_name = ""
        self.partition_timelimits: dict = {}

    def get_columns(self) -> List[str]:
        """The list of columns requested from inquirer.

        Returns:
            The current columns of this collection.
        """
        return self.columns

    def get_jobs(self) -> List[str]:
        """List of jobs to get from inquirer.

        Returns:
            job names as a sorted list
        """
        return sorted({job.job for job in self.jobs.values()})

    def set_out_dir(self, directory: str) -> None:
        """Set this collection's directory to try parsing out jobs from output files.

        If directory is empty, will use the current directory, otherwise will
        try to get the full path name.

        Args:
            directory: the directory to search for job names.

        Raises:
            ValueError: if the directory does not exist
            ValueError: if the directory contains no files
            ValueError: if the directory contains no valid files
        """
        # set and validate working directory to full path
        if directory == "":
            working_directory = os.getcwd()
        else:
            working_directory = os.path.realpath(directory)

        if not os.path.exists(working_directory):
            raise ValueError(f"{working_directory} does not exist!")

        # get files from directory
        files = os.listdir(working_directory)
        files = list(
            filter(lambda x: os.path.isfile(os.path.join(working_directory, x)), files)
        )
        if len(files) == 0:
            raise ValueError(f"{working_directory} contains no files!")

        for file in files:
            self.process_seff_file(file)

        if len(self.jobs) == 0:
            raise ValueError(f"{working_directory} contains no valid output files!")
        self.dir_name = working_directory

    def set_jobs(self, jobs: tuple) -> None:
        """Set the collection jobs to the provided job ids.

        if jobs is empty, use the current working directory
        if jobs is a singleton and a directory, set to that directory
        else, parse jobs as files or raw job ids

        Args:
            jobs: Tuple of job names to use for populating the collection

        Raises:
            ValueError: if no valid jobs are provided
        """
        if jobs == ():
            # look in current directory for slurm outputs
            self.set_out_dir("")
            return
        if len(jobs) == 1 and os.path.isdir(jobs[0]):
            self.set_out_dir(jobs[0])
            return
        for job_id in jobs:
            match = self.job_regex.match(job_id)

            if match:
                self.add_job(match.group("job"), match.group("jobid"))
            else:
                self.process_seff_file(job_id)

        if len(self.jobs) == 0:
            raise ValueError("No valid jobs provided!")

    def process_seff_file(self, filename: str) -> None:
        """Try to parse out job information from the supplied filename.

        Args:
            filename: the filename to try and match
        """
        match = self.job_file_regex.match(filename)
        if match:
            self.add_job(match.group("job"), match.group("jobid"), filename)

    def add_job(self, job: str, jobid: str, filename: Optional[str] = None) -> None:
        """Add a job to the collection.

        Args:
            job: the 'base' job number
            jobid: equal to the job unless it is an array job
            filename: the filename of the out file this job is derived from
        """
        self.jobs[jobid] = Job(job, jobid, filename)

    def process_entry(self, entry: Dict, add_job: bool = False) -> None:
        """Update the jobs collection with information from the provided entry.

        Args:
            entry: the account entry from a db inquirer
            add_job: if true, will add the job to the collection if it doesn't exist
        """
        job_id = entry["JobID"].split(".")[0]
        job_id_raw = entry["JobIDRaw"].split(".")[0]
        if job_id not in self.jobs:
            match = self.job_regex.match(job_id)
            # job is in jobs
            if match and (match.group("job") in self.jobs or add_job):
                self.add_job(match.group("job"), job_id)
            # check if the job_id is an array job
            elif job_id_raw in self.jobs:
                old_job = self.jobs.pop(job_id_raw)
                self.add_job(old_job.job, job_id, old_job.filename)
            else:
                return

        # handle partition limit for timelimit
        if (
            "Timelimit" in entry
            and entry["Timelimit"] == "Partition_Limit"
            and "Partition" in entry
            and entry["Partition"] in self.partition_timelimits
        ):
            entry["Timelimit"] = self.partition_timelimits[entry["Partition"]]

        self.jobs[job_id].update(entry)

    def get_sorted_jobs(self, change_sort: bool) -> List[Job]:
        """Sort the jobs.

        Args:
            change_sort: a switch to indicate the type of sorting
                if true will sort by the modified time of the file or job number
                if false will sort by the if the file exists,
                the length, then the content

        Returns:
            sorted list of jobs to display
        """

        def get_time(job: Job) -> float:
            # handle None and '', use numeric representation of name
            idnum = float(re.sub("[^0-9.]", "", job.jobid.replace("_", ".")))
            file = job.filename
            if file and self.dir_name:
                file = os.path.join(self.dir_name, file)
            if file and os.path.exists(file):
                return os.path.getmtime(file)
            return idnum

        def get_file_name(job: Job) -> Tuple[bool, int, str]:
            file = job.name()
            file = os.path.join(self.dir_name, file)
            return (not os.path.exists(file), len(file), file)

        if change_sort:
            return sorted(self.jobs.values(), key=get_time, reverse=True)

        return sorted(self.jobs.values(), key=get_file_name)

    def set_partition_limits(self, limits: dict) -> None:
        """Set partition limits from db inquirer.

        Args:
            limits: dict of partition to partition timelimit
        """
        self.partition_timelimits = limits



"""Module for rendering tabulated values."""
import copy
import re
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Union

import click

from .job import Job, state_colors


#: Regex for format tokens, NAME[%[ALIGNMENT][WIDTH[e?]]]
FORMAT_RE = re.compile(
    r"(?P<title>[^%:]+)"  # must have title
    r"([%:]"  # may have other formatting
    r"(?P<alignment>[<^>])?"  # like alignment
    r"((?P<width>\d+)(?P<end>[e$])?)?"  # if width is present, may have an e
    r")?"
)


class OutputRenderer:
    """A collection of formatting columns for rendering output."""

    def __init__(
        self,
        valid_titles: List,
        format_str: str = "JobID%>,State,Elapsed%>,CPUEff,MemEff",
        node: bool = False,
        gpu: bool = False,
        parsable: bool = False,
    ) -> None:
        """Initialize renderer with format string and list of valid titles.

        Args:
            valid_titles: List of valid options for format tokens
            format_str: comma separated list of format tokens
            node: bool indicating if node-level data should be reported
            gpu: in addition to node, should each GPU be reported
            parsable: should output be rendered in a parsable format
        """
        self.node = node
        self.gpu = gpu
        # values required for proper parsing, need not be included in output
        self.required = ["JobID", "JobIDRaw", "State", "AdminComment"]
        # values derived from other values, list includes all dependent values
        self.derived: Dict[str, List] = {
            "CPUEff": ["TotalCPU", "AllocCPUS", "Elapsed"],
            "MemEff": ["REQMEM", "NNodes", "AllocCPUS", "MaxRSS"],
            "TimeEff": ["Elapsed", "Timelimit"],
            "GPU": [],
            "GPUMem": [],
            "GPUEff": [],
        }

        self.parsable = parsable

        # build formatters
        self.formatters = build_formatters(format_str)

        # validate with titles and derived keys
        valid_titles += self.derived.keys()
        self.query_columns = self.validate_formatters(valid_titles)

        # build columns for sacct call
        self.correct_columns()

    def validate_formatters(self, valid_titles: List) -> List:
        """Validate titles of formatters attribute.

        Expands GPU to GPUEff and GPUMem in formatters

        Args:
            valid_titles: List of valid options for format tokens

        Returns:
            Return list of validated strings to query
        """
        result = [fmt.validate_title(valid_titles) for fmt in self.formatters]

        if self.node:
            if "JobID" not in self.formatters:
                self.formatters.insert(0, ColumnFormatter("JobID"))
            # ensure alignment is <, regardless of inputs
            self.formatters[self.formatters.index("JobID")].alignment = "<"

        if "GPU" in self.formatters:
            ind = self.formatters.index("GPU")
            self.formatters[ind].title = "GPUEff"
            gpu_mem = copy.copy(self.formatters[ind])
            gpu_mem.title = "GPUMem"
            gpu_mem.color_function = lambda x: render_eff(x, "mid")
            self.formatters.insert(ind + 1, gpu_mem)

        if (
            self.gpu
            and "GPUEff" not in self.formatters
            and "GPUMem" not in self.formatters
        ):
            # need to assign color functions as that normally happens before this
            formatter = ColumnFormatter("GPUEff")
            self.formatters.append(formatter)
            formatter = ColumnFormatter("GPUMem")
            self.formatters.append(formatter)

        return result

    def correct_columns(self) -> None:
        """Expand derived values of query columns and remove duplicates."""
        result: List[List] = [
            self.derived[c] if c in self.derived else [c] for c in self.query_columns
        ]
        # flatten
        flat_result = [item for sublist in result for item in sublist]

        # add in required values
        flat_result += self.required

        # remove duplicates
        self.query_columns = list(sorted(set(flat_result)))

    def format_jobs(self, jobs: List[Job]) -> str:
        """Given list of jobs, build output table.

        Args:
            jobs: List of job objects

        Returns:
            Formatted table as single string
        """
        result = ""
        delimiter = "|" if self.parsable else " "

        if len(self.formatters) == 0:
            return result

        if len(self.formatters) == 1:
            # if only one formatter is present, override the alignment
            self.formatters[0].no_formatting()
            # skip adding the title to result

        else:
            for fmt in self.formatters:
                if self.parsable:
                    fmt.no_formatting()
                else:
                    fmt.compute_width(jobs, self.node, self.gpu)

            result += delimiter.join(
                fmt.format_title(bold=not self.parsable) for fmt in self.formatters
            )
            if len(jobs) != 0:
                result += "\n"

        if self.node:
            # join each row by newlines
            result += "\n".join(
                # join each column entry by spaces
                delimiter.join(str(column) for column in columns).rstrip()
                # for each job
                for job in jobs
                # columns is a tuple of generators from format_node_job
                for columns in zip(
                    *(fmt.format_node_job(job, self.gpu) for fmt in self.formatters)
                )
            )

        else:
            result += "\n".join(
                delimiter.join(fmt.format_job(job) for fmt in self.formatters).rstrip()
                for job in jobs
            )

        return result


class ColumnFormatter:
    """A single column formatting object."""

    def __init__(self, token: str) -> None:
        """Build column entry.

        Args:
            token: format string of the form NAME[%[ALIGNMENT][WIDTH[e?]]]

        Raises:
            ValueError: if unable to parse the format token
        """
        match = re.fullmatch(FORMAT_RE, token)
        if not match or (
            ("%" in token or ":" in token)
            and not match.group("alignment")
            and not match.group("width")
        ):
            err = f"Unable to parse format token '{token}'"
            if "%" in token:
                err += ", did you forget to wrap in quotes?"
            raise ValueError(err)

        self.title = match.group("title")

        self.alignment = match.group("alignment")
        if not self.alignment:
            self.alignment = "^"

        self.width = None
        if match.group("width"):
            self.width = int(match.group("width"))  # none will be calculated later

        self.end = match.group("end")

        self.color_function: Callable[[str], Tuple[str, Any]] = lambda x: (str(x), None)
        fold_title = self.title.casefold()
        if fold_title == "state":
            self.color_function = color_state
        elif fold_title in ("cpueff", "gpueff", "gpu"):
            self.color_function = lambda x: render_eff(x, "high")
        elif fold_title in ("timeeff", "memeff", "gpumem"):
            self.color_function = lambda x: render_eff(x, "mid")

    def __eq__(self, other: Any) -> bool:
        """Test for equality.

        Args:
            other: other object

        Returns:
            True if the other object is a column formatter with matching attributes
            or if this title match the other object as a string

        """
        if isinstance(other, ColumnFormatter):
            for k in self.__dict__:
                if k != "color_function" and self.__dict__[k] != other.__dict__[k]:
                    return False
            return True
        if isinstance(other, str):
            return self.title == other
        return False

    def __repr__(self) -> str:
        """Recreate format string of formatter.

        Returns:
            String representation of column formatter
        """
        return f"{self.title}%{self.alignment}{self.width}"

    def validate_title(self, valid_titles: List[str]) -> str:
        """Validate the title against a list.

        Tries to find this formatter's title in the valid titles list in a case
        insensitive manner.  If found, replace with valid_title to correct
        capitalization to match valid_titles entry.

        Args:
            valid_titles: list of valid title strings

        Returns:
            The self title validated from the valid_titles list

        Raises:
            ValueError: if self.title is not found in the valid list
        """
        fold_title = self.title.casefold()
        for title in valid_titles:
            if fold_title == title.casefold():
                self.title = title
                return title

        raise ValueError(
            f"'{self.title}' is not a valid title. "
            "Run sacct --helpformat for a list of allowed values."
        )

    def compute_width(
        self,
        jobs: List[Job],
        node: bool = False,
        gpu: bool = False,
    ) -> None:
        """Set width for this column based on job listing.

        Determine the max width of all entries if the width attribute is unset.
        Includes title in determination

        Args:
            jobs: List of job objects to consider
            node: If True, report individual node stats
            gpu: If True, report individual gpu stats
        """
        if self.width is not None:
            return

        self.width = len(self.title)
        if len(jobs) > 0:
            if node:
                width = max(
                    len(str(entry))
                    for job in jobs
                    for entry in job.get_node_entries(self.title, gpu)
                )
            else:
                width = max(len(str(job.get_entry(self.title))) for job in jobs)
            self.width = max(self.width, width)

        self.width += 2  # add some boarder

    def no_formatting(self) -> None:
        """Set the formatter to just display the entries."""
        self.alignment = "<"
        self.width = None
        self.color_function = lambda x: (str(x), None)

    def format_title(self, bold: bool = True) -> str:
        """Format title of column for printing.

        Args:
            bold: if true, the resulting string will be styled bold

        Returns:
            the formatted title
        """
        result = self.format_entry(self.title)
        return click.style(result, bold=bold)

    def format_job(self, job: Job) -> str:
        """Format the provided job for printing.

        Args:
            job: the Job to format

        Returns:
            the formatted job entry
        """
        value = job.get_entry(self.title)
        return self.format_entry(*self.color_function(value))

    def format_node_job(
        self, job: Job, gpu: bool = False
    ) -> Generator[str, None, None]:
        """Format the provided job for printing with individual node stats.

        Args:
            job: the Job to format
            gpu: If True, report individual gpu stats

        Returns:
            generator with values for the job
        """
        return (
            self.format_entry(*self.color_function(value))
            for value in job.get_node_entries(self.title, gpu)
        )

    def format_entry(self, entry: str, color: Optional[str] = None) -> str:
        """Format the entry to match width, alignment, and color.

        If no color is supplied, will just return string
        If supplied, use click.style to change fg
        If the entry is longer than self.width, truncate the end

        Args:
            entry: the string to format
            color: set foreground of style string

        Returns:
            entry colored, aligned, and possibly truncated
        """
        if self.width is None:
            result = entry
        else:
            if self.end:
                entry = entry[-self.width :]
            else:
                entry = entry[: self.width]
            result = (f"{{:{self.alignment}{self.width}}}").format(entry)

        if color:
            result = click.style(result, fg=color)
        return result


def color_state(value: str) -> Tuple[str, Optional[str]]:
    """Get the color name of the provided state string.

    Args:
        value: the state name of the job

    Returns:
        The state value and it's color (if found) or None
    """
    return value, state_colors.get(value, None)


def render_eff(value: Union[str, float], target_type: str) -> Tuple[str, Optional[str]]:
    """Return a styled string for efficiency values.

    Args:
        value: the number or string to render
        target_type: "mid" or "high", the type of color map to use

    Returns:
        A tuple with:
        the value formatted with a percent sybol
        the color or None if it should remain the default
    """
    color_maps = {"mid": color_mid, "high": color_high}
    if isinstance(value, str):  # a "---"
        color = None
    else:
        color = color_maps[target_type](value)
        value = f"{value}%"
    return value, color


def color_mid(value: float) -> Optional[str]:
    """Determine color for efficiency value where "mid" values are the target.

    Args:
        value: percent (0-100) of value to color

    Returns:
        The color string for click or None if color should be unchanged
    """
    if value < 20 or value > 90:  # too close to limit
        return "red"
    if value > 60:  # good
        return "green"
    return None


def color_high(value: float) -> Optional[str]:
    """Determine color for efficiency value where "high" values are the target.

    Args:
        value: percent (0-100) of value to color

    Returns:
        The color string for click or None if color should be unchanged
    """
    if value < 20:  # too low
        return "red"
    if value > 80:  # good
        return "green"
    return None


def build_formatters(format_str: str) -> List:
    """Generate list of formatters from comma separated list in format string.

    Args:
        format_str: comma separated list of format tokens

    Returns:
        Return list of ColumnFormatters
    """
    return [ColumnFormatter(fmt) for fmt in format_str.split(",") if fmt != ""]



"""CLI for reportseff."""
from shutil import which
import sys
from typing import Any, Dict, List, Tuple

import click

from . import __version__
from .db_inquirer import BaseInquirer, SacctInquirer
from .job_collection import JobCollection
from .output_renderer import OutputRenderer
from .parameters import ReportseffParameters


@click.command()
@click.option(
    "--modified-sort",
    default=False,
    is_flag=True,
    help="If set, will sort outputs by modified time of files",
)
@click.option(
    "--color/--no-color",
    default=True,
    help="Force color output. No color will use click defaults",
)
@click.option(
    "--format",
    "format_str",
    default="JobID%>,State,Elapsed%>,TimeEff,CPUEff,MemEff",
    help="Comma-separated list of columns to include. Options "
    "are any valid sacct input along with CPUEff, MemEff, and "
    "TimeEff.  In systems with jobstat caching, GPU usage can be "
    "added with GPUEff, GPUMem or GPU (for both). "
    "A width and alignment may optionally be provided "
    'after "%", e.g. JobID%>15 aligns job id right with max '
    "width of 15 characters. Generally NAME[[%:][ALIGNMENT][WIDTH[e$]?]]. "
    "When an `e` or `$` is present after a width argument, "
    "the output will be truncated to the right."
    "Prefix with a + to add to the defaults. "
    "A single format token will suppress the header line. "
    "Wrap in quotes to pass a string literal, "
    "otherwise alignment may be misinterpreted.",
)
@click.option(
    "--debug", default=False, is_flag=True, help="Print raw db query to stderr"
)
@click.option(
    "-u",
    "--user",
    default="",
    help="Ignore jobs, return all jobs in last week from user",
)
@click.option(
    "--partition",
    default="",
    help="Only include jobs with the specified partition",
)
@click.option(
    "-s", "--state", default="", help="Only include jobs with the specified states"
)
@click.option(
    "-S", "--not-state", default="", help="Include jobs without the specified states"
)
@click.option(
    "--since",
    default="",
    help="Only include jobs before this time. Can be valid sacct "
    "or as a comma separated list of time deltas, e.g. d=2,h=1 "
    "means 2 days, 1 hour before current time. Weeks, days, "
    "hours, and minutes can use case-insensitive abbreviations. "
    "Minutes is the minimum resolution, while weeks is the coarsest.",
)
@click.option(
    "--until",
    default="",
    help="Only include jobs before this time. Can be valid sacct "
    "or as a comma separated list of time deltas, e.g. d=2,h=1 "
    "means 2 days, 1 hour before current time. Weeks, days, "
    "hours, and minutes can use case-insensitive abbreviations. "
    "Minutes is the minimum resolution, while weeks is the coarsest.",
)
@click.option(
    "--node/--no-node",
    "-n/-N",
    default=False,
    help="Report node-level statistics. Adds `jobid` to format for proper display.",
)
@click.option(
    "--node-and-gpu/--no-node-gpu",
    "-g/-G",
    default=False,
    help=(
        "Report each GPU for each node. "
        "Sets `node` and adds `GPU` to format automatically."
    ),
)
@click.option(
    "--parsable",
    "-p",
    is_flag=True,
    default=False,
    help="Output will be '|' delmited without a '|' at the end.",
)
@click.version_option(version=__version__)
@click.argument("jobs", nargs=-1)
def main(**kwargs: Any) -> None:
    """Main entry point for reportseff."""
    args = ReportseffParameters(**kwargs)

    output, entries = get_jobs(args)

    if entries > 20:
        click.echo_via_pager(output, color=args.color)
    else:
        click.echo(output, color=args.color)


def get_jobs(args: ReportseffParameters) -> Tuple[str, int]:
    """Helper method to get jobs from db_inquirer.

    Returns:
        The string to display, tabulated and colored
        The number of jobs found to use paging properly

    Raises:
        Exception: if there is an error processing entries
    """
    job_collection = JobCollection()

    inquirer, renderer = get_implementation(
        args.format_str, args.node, args.node_and_gpu, args.parsable
    )

    inquirer.set_state(args.state)
    inquirer.set_not_state(args.not_state)

    inquirer.set_since(args.since)
    inquirer.set_until(args.until)

    inquirer.set_partition(args.partition)
    add_jobs = False

    try:
        if args.user:
            inquirer.set_user(args.user)
            add_jobs = True
        elif inquirer.has_since() and not args.jobs:  # since is set
            inquirer.all_users()
            add_jobs = True
        else:
            job_collection.set_jobs(args.jobs)

    except ValueError as error:
        click.secho(str(error), fg="red", err=True)
        sys.exit(1)

    job_collection.set_partition_limits(inquirer.get_partition_timelimits())
    db_output = get_db_output(inquirer, renderer, job_collection, args.debug)
    for entry in db_output:
        try:
            job_collection.process_entry(entry, add_job=add_jobs)
        except Exception as error:
            click.echo(f"Error processing entry: {entry}", err=True)
            raise error

    found_jobs = job_collection.get_sorted_jobs(args.modified_sort)
    found_jobs = [j for j in found_jobs if j.state]

    return renderer.format_jobs(found_jobs), len(found_jobs)


def get_implementation(
    format_str: str,
    node: bool = False,
    node_and_gpu: bool = False,
    parsable: bool = False,
) -> Tuple[BaseInquirer, OutputRenderer]:
    """Get system-specific objects.

    Args:
        format_str: the formatting options specified by user

    Returns:
        A db_inqurirer
        An output renderer
    """
    if which("sacct") is not None:
        inquirer = SacctInquirer()
        renderer = OutputRenderer(
            inquirer.get_valid_formats(),
            format_str,
            node=node or node_and_gpu,
            gpu=node_and_gpu,
            parsable=parsable,
        )
    else:
        click.secho("No supported scheduling systems found!", fg="red", err=True)
        sys.exit(1)

    return inquirer, renderer


def get_db_output(
    inquirer: BaseInquirer,
    renderer: OutputRenderer,
    job_collection: JobCollection,
    debug: bool,
) -> List[Dict[str, str]]:
    """Get output from inquirer.

    Returns:
        The db inquirer entries for the provided objects
    """

    def print_debug(info: str) -> None:
        click.echo(info, err=True)

    debug_cmd = None
    if debug:
        debug_cmd = print_debug

    try:
        result = inquirer.get_db_output(
            renderer.query_columns, job_collection.get_jobs(), debug_cmd
        )
    except RuntimeError as error:
        click.secho(str(error), fg="red", err=True)
        sys.exit(1)

    return result



"""Abstract and concrete implementations of scheduler databases."""
from abc import ABC, abstractmethod
import datetime
import re
import subprocess
from typing import Callable, Dict, List, Optional, Set

import click


class BaseInquirer(ABC):
    """Abstract interface for inquiring different schedulers."""

    def __init__(self) -> None:
        """Initialize a new inquirer."""

    @abstractmethod
    def get_valid_formats(self) -> List[str]:
        """Get the valid formatting options supported by the inquirer.

        Returns:
            List of valid format options
        """

    @abstractmethod
    def set_sacct_args(self, jobs: List[str]) -> List[str]:
        """Set arguments of sacct query.

        Args:
            jobs: list of job names

        Returns:
            String of sacct arguments

        Raises:
            RuntimeError: if sacct doesn't return properly
        """

    @abstractmethod
    def get_db_output(
        self,
        columns: List[str],
        jobs: List[str],
        debug_cmd: Optional[Callable],
    ) -> List[Dict[str, str]]:
        """Query the database with the supplied columns.

        Args:
            columns: validated format names as strings
            jobs: list of job names
            debug_cmd: If specified, the raw output will passed to this function

        Returns:
            List of rows, where each row is a dictionary
            with the columns as keys and entries as values
            Output order is not garunteed to match the jobs list

        """

    @abstractmethod
    def set_user(self, user: str) -> None:
        """Set the collection of jobs based on the provided user.

        Args:
            user: user name
        """

    @abstractmethod
    def set_partition(self, partition: str) -> None:
        """Set the collection of jobs based on the provided partition.

        Args:
            partition: partition name
        """

    @abstractmethod
    def all_users(self) -> None:
        """Ignore provided jobs, query for all users."""

    @abstractmethod
    def set_state(self, state: str) -> None:
        """Set the state to include output jobs.

        Args:
            state: comma separated list of state names or codes
        """

    @abstractmethod
    def set_not_state(self, state: str) -> None:
        """Set the state to exclude from output jobs.

        Args:
            state: comma separated list of state names or codes
        """

    @abstractmethod
    def parse_date(self, d: str) -> str:
        """Parse and convert custom string date format.

        Args:
            d: the string of date.

        Returns:
            converted string of date
        """

    @abstractmethod
    def set_until(self, until: str) -> None:
        """Set the filter for time of jobs to consider.

        Args:
            until: the string for filtering.  If specified as time=amount
                will subtract that amount from the current time
        """

    @abstractmethod
    def set_since(self, since: str) -> None:
        """Set the filter for time of jobs to consider.

        Args:
            since: the string for filtering.  If specified as time=amount
                will subtract that amount from the current time
        """

    @abstractmethod
    def has_since(self) -> bool:
        """Tests if `since` has been set.

        Returns:
            True if set_since has been called on this inquirer
        """

    @abstractmethod
    def get_partition_timelimits(self) -> dict:
        """Get partition time limits.

        Returns:
            dict mapping partition names to maximum timelimits.
        """


class SacctInquirer(BaseInquirer):
    """Implementation of BaseInquirer for the sacct slurm function."""

    def __init__(self) -> None:
        """Initialize a new inquirer."""
        self.default_args = "sacct -P -n".split()
        self.user: Optional[str] = None
        self.state: Optional[Set] = None
        self.not_state: Optional[Set] = None
        self.since: Optional[str] = None
        self.until: Optional[str] = None
        self.query_all_users: bool = False
        self.partition: Optional[str] = None

    def get_valid_formats(self) -> List[str]:
        """Get the valid formatting options supported by the inquirer.

        Returns:
            List of valid format options

        Raises:
            RuntimeError: if sacct raises an error
        """
        command_args = "sacct --helpformat".split()
        cmd_result = subprocess.run(
            args=command_args,
            stdout=subprocess.PIPE,
            encoding="utf8",
            check=True,
            universal_newlines=True,
            shell=False,
        )
        if cmd_result.returncode != 0:
            raise RuntimeError("Error retrieving sacct options with --helpformat")
        result = cmd_result.stdout.split()
        return result

    def set_sacct_args(self, jobs: List[str]) -> List[str]:
        """Set arguments of sacct query.

        Args:
            jobs: list of job names

        Returns:
            String of sacct arguments

        """
        args = []
        if self.user:
            if not self.since:
                start_date = datetime.date.today() - datetime.timedelta(days=7)
                self.since = start_date.strftime("%m%d%y")  # MMDDYY
            args += [f"--user={self.user}", f"--starttime={self.since}"]
        elif self.query_all_users:
            args += ["--allusers", f"--starttime={self.since}"]
        else:
            args += ["--jobs=" + ",".join(jobs)]
            if self.since:
                args += [f"--starttime={self.since}"]

        if self.partition:
            args += [f"--partition={self.partition}"]
        if self.until:
            args += [f"--endtime={self.until}"]
        return args

    def get_db_output(
        self,
        columns: List[str],
        jobs: List[str],
        debug_cmd: Optional[Callable] = None,
    ) -> List[Dict[str, str]]:
        """Query the database with the supplied columns.

        Args:
            columns: validated format names as strings
            jobs: list of job names
            debug_cmd: If specified, the raw output will passed to this function

        Returns:
            List of rows, where each row is a dictionary
            with the columns as keys and entries as values
            Output order is not guaranteed to match the jobs list

        Raises:
            RuntimeError: if sacct doesn't return properly
        """
        args = self.default_args + ["--format=" + ",".join(columns)]
        args += self.set_sacct_args(jobs)
        try:
            cmd_result = subprocess.run(
                args=args,
                stdout=subprocess.PIPE,
                encoding="utf8",
                check=True,
                universal_newlines=True,
                shell=False,
            )
            cmd_result.check_returncode()

        except subprocess.CalledProcessError as error:
            raise RuntimeError(f"Error running sacct!\n{error.stderr}") from error

        lines = cmd_result.stdout.split("\n")
        if debug_cmd is not None:
            debug_cmd("\n".join(lines))

        result = [dict(zip(columns, line.split("|"))) for line in lines if line]

        if self.state:
            # split to get first word in entries like "CANCELLED BY X"
            result = [r for r in result if r["State"].split()[0] in self.state]

        if self.not_state:
            # split to get first word in entries like "CANCELLED BY X"
            result = [r for r in result if r["State"].split()[0] not in self.not_state]

        return result

    def set_user(self, user: str) -> None:
        """Set the collection of jobs based on the provided user.

        Args:
            user: user name
        """
        self.user = user

    def set_partition(self, partition: str) -> None:
        """Set the collection of jobs based on the provided partition.

        Args:
            partition: partition name
        """
        self.partition = partition

    def all_users(self) -> None:
        """Query for all users if `since` is set."""
        self.query_all_users = True

    def set_state(self, state: str) -> None:
        """Set the state to include output jobs.

        Args:
            state: comma separated list of state names or codes
        """
        if not state:
            return

        self.state = get_states_as_set(state)
        # add a single value if it's empty here
        if not self.state:
            click.secho("No valid states provided to include", fg="yellow", err=True)
            self.state.add(None)

    def set_not_state(self, state: str) -> None:
        """Set the state to exclude from output jobs.

        Args:
            state: comma separated list of state names or codes

        """
        if not state:
            return

        self.not_state = get_states_as_set(state)
        # add a single value if it's empty here
        if not self.not_state:
            click.secho("No valid states provided to exclude", fg="yellow", err=True)
            self.not_state = None

    def parse_date(self, d: str) -> str:
        """Parse and convert custom string date format.

        Args:
            d: the string of date.

        Returns:
            converted string of date
        """
        abbrev_to_key = {
            "w": "weeks",
            "W": "weeks",
            "d": "days",
            "D": "days",
            "h": "hours",
            "H": "hours",
            "m": "minutes",
            "M": "minutes",
        }
        valid_args = ["weeks", "days", "hours", "minutes"]
        date_args = {}

        args = d.split(",")
        for arg in args:
            toks = arg.split("=")

            # lines don't have an equal
            if len(toks) < 2:
                continue

            # convert key to name
            if toks[0] in abbrev_to_key:
                toks[0] = abbrev_to_key[toks[0]]

            toks[0] = toks[0].lower()

            if toks[0] in valid_args:
                try:
                    date_args[toks[0]] = int(toks[1])
                except ValueError:
                    continue

        date = datetime.datetime.today()
        date -= datetime.timedelta(**date_args)
        return date.strftime("%Y-%m-%dT%H:%M")  # MMDDYY

    def set_until(self, until: str) -> None:
        """Set the filter for time of jobs to consider.

        Args:
            until: the string for filtering. If specified as time=amount
                will subtract that amount from the current time
        """
        if not until:
            return
        if "=" in until:  # handle custom format
            self.until = self.parse_date(until)
        else:
            self.until = until

    def set_since(self, since: str) -> None:
        """Set the filter for time of jobs to consider.

        Args:
            since: the string for filtering.  If specified as time=amount
                will subtract that amount from the current time
        """
        if not since:
            return
        if "=" in since:  # handle custom format
            self.since = self.parse_date(since)
        else:
            self.since = since

    def has_since(self) -> bool:
        """Check if since has been set.

        Returns:
            True if since has been set properly
        """
        return bool(self.since)

    def get_partition_timelimits(self) -> dict:
        """Get partition time limits.

        Returns:
            dict mapping partition names to maximum timelimits.

        Raises:
            RuntimeError: if scontrol raises an error
        """
        command_args = "scontrol show partition".split()
        cmd_result = subprocess.run(
            args=command_args,
            stdout=subprocess.PIPE,
            encoding="utf8",
            check=True,
            universal_newlines=True,
            shell=False,
        )
        if cmd_result.returncode != 0:
            raise RuntimeError("Error retrieving information from scontrol")

        partition_name = re.compile(r"^PartitionName=(?P<name>\S+)$")
        time_limit = re.compile(r"MaxTime=(?P<time>\S+)")

        partition = ""
        result = {}
        for line in cmd_result.stdout.split():
            match = re.match(partition_name, line)
            if match:
                partition = match.group("name")
            match = re.match(time_limit, line)
            if match:
                result[partition] = match.group("time")

        return result


def get_states_as_set(state_list: str) -> Set:
    """Helper method to parse the state string.

    Args:
        state_list: comma separated string with codes and states

    Returns:
        Set with valid state names in upper case
    """
    codes_to_states = {
        "BF": "BOOT_FAIL",
        "CA": "CANCELLED",
        "CD": "COMPLETED",
        "DL": "DEADLINE",
        "F": "FAILED",
        "NF": "NODE_FAIL",
        "OOM": "OUT_OF_MEMORY",
        "PD": "PENDING",
        "PR": "PREEMPTED",
        "R": "RUNNING",
        "RQ": "REQUEUED",
        "RS": "RESIZING",
        "RV": "REVOKED",
        "S": "SUSPENDED",
        "TO": "TIMEOUT",
    }
    possible_states = set(codes_to_states.values())
    states = {
        codes_to_states.get(state, state) for state in state_list.upper().split(",")
    }

    for state in states:
        if state not in possible_states:
            click.secho(f"Unknown state {state}", fg="yellow", err=True)

    return states.intersection(possible_states)



"""Module for representing scheduler jobs."""
import base64
from datetime import timedelta
import gzip
import json
import re
from typing import Any, Dict, Generator, Optional, Union


multiple_map = {
    "K": 1024**0,
    "M": 1024**1,
    "G": 1024**2,
    "T": 1024**3,
    "E": 1024**4,
}

state_colors = {
    "FAILED": "red",
    "TIMEOUT": "red",
    "OUT_OF_MEMORY": "red",
    "RUNNING": "cyan",
    "CANCELLED": "yellow",
    "COMPLETED": "green",
    "PENDING": "blue",
}

#: Regex for DDHHMMSS style timestamps
DDHHMMSS_RE = re.compile(
    r"(?P<days>\d+)-(?P<hours>\d{2}):(?P<minutes>\d{2}):(?P<seconds>\d{2})"
)
#: Regex for HHMMSS style timestamps
HHMMSS_RE = re.compile(r"(?P<hours>\d{2}):(?P<minutes>\d{2}):(?P<seconds>\d{2})")
#: Regex for HHMMmmm style timestamps
MMSSMMM_RE = re.compile(
    r"(?P<minutes>\d{2}):(?P<seconds>\d{2}).(?P<milliseconds>\d{3})"
)
#: Regex for maxRSS and reqmem
MEM_RE = re.compile(
    r"(?P<memory>[-+]?\d*\.\d+|\d+)(?P<multiple>[KMGTE]?)(?P<type>[nc]?)"
)


class Job:
    """Representation of scheduler job."""

    def __init__(self, job: str, jobid: str, filename: Optional[str]) -> None:
        """Initialize new job.

        Args:
            job: the base job number
            jobid: same as job unless an array job
            filename: the output file associated with this job
        """
        self.job = job
        self.jobid = jobid
        self.filename = filename
        self.stepmem = 0.0
        self.totalmem: Optional[float] = None
        self.time: Optional[str] = "---"
        self.time_eff: Union[str, float] = "---"
        self.cpu: Optional[Union[str, float]] = "---"
        self.mem: Union[str, float] = "---"
        self.state: Optional[str] = None
        self.mem_eff: Optional[float] = None
        self.gpu: Optional[float] = None
        self.gpu_mem: Optional[float] = None
        self.other_entries: Dict[str, Any] = {}
        # safe to cache now
        self.other_entries["JobID"] = self.name()
        self.comment_data: Dict = {}

    def __eq__(self, other: Any) -> bool:
        """Test for equality.

        Args:
            other: the other object

        Returns:
            true if the other object is a Job and all attributes match
        """
        if not isinstance(other, Job):
            return False

        return self.__dict__ == other.__dict__

    def __repr__(self) -> str:
        """Job representation.

        Returns:
            The job string representation
        """
        return f"Job(job={self.job}, jobid={self.jobid}, filename={self.filename})"

    def update(self, entry: Dict) -> None:
        """Update the job properties based on the db_inquirer entry.

        Args:
            entry: the db_inquirer entry for the matching job
        """
        if "." not in entry["JobID"]:
            self.state = entry["State"].split()[0]

        if self.state == "PENDING":
            self._cache_entries()
            return

        # main job id
        if self.jobid == entry["JobID"]:
            self._update_main_job(entry)
            self._cache_entries()

        elif self.state != "RUNNING":
            for k, value in entry.items():
                if k not in self.other_entries or not self.other_entries[k]:
                    self.other_entries[k] = value
            self.stepmem += parsemem(entry["MaxRSS"]) if "MaxRSS" in entry else 0

    def _update_main_job(self, entry: Dict) -> None:
        """Update properties for the main job.

        Args:
            entry: the entry where the jobid matches exactly, e.g. not batch or ex
        """
        for k, value in entry.items():
            if k not in self.other_entries or not self.other_entries[k]:
                self.other_entries[k] = value
        self.time = entry["Elapsed"] if "Elapsed" in entry else None

        requested = 0
        if "Timelimit" in entry and entry["Timelimit"] not in (
            "UNLIMITED",
            "Partition_Limit",
        ):
            requested = _parse_slurm_timedelta(entry["Timelimit"])

        wall = _parse_slurm_timedelta(entry["Elapsed"]) if "Elapsed" in entry else 0

        if requested != 0:
            self.time_eff = round(wall / requested * 100, 1)

        if self.state == "RUNNING":
            return

        cpus = (
            _parse_slurm_timedelta(entry["TotalCPU"]) / int(entry["AllocCPUS"])
            if "TotalCPU" in entry and "AllocCPUS" in entry
            else 0
        )
        if wall == 0:
            self.cpu = None
        else:
            self.cpu = round(cpus / wall * 100, 1)

        if "REQMEM" in entry and "NNodes" in entry and "AllocCPUS" in entry:
            self.totalmem = parsemem(
                entry["REQMEM"], int(entry["NNodes"]), int(entry["AllocCPUS"])
            )

        if "AdminComment" in entry and len(entry["AdminComment"]) > 10:
            self._parse_admin_comment(entry["AdminComment"])

    def _parse_admin_comment(self, comment: str) -> None:
        """Use admin command to override efficiency values.

        Decodes and parses admincommand from jobstats.

        Args:
            comment: The AdminComment field.

        Raises:
            ValueError: if the comment doesn't start with JS1.
            ValueError: if the comment can't be decoded.
        """
        comment_type = comment[:3]

        if not comment_type.startswith("JS"):
            # ignore comments that aren't from jobstats (JS)
            return

        if comment_type not in ("JS1",):
            raise ValueError(f"Unknown comment type '{comment_type}'")
        data = {}
        try:
            data = json.loads(gzip.decompress(base64.b64decode(comment[4:])))
        except BaseException as exception:
            raise ValueError(f"Cannot decode comment '{comment}'") from exception

        def average(value: str, data: Optional[dict] = None) -> float:
            if data is None:
                data = self.comment_data
            return round(
                sum(v[value] for v in data.values()) / len(data),
                1,
            )

        for node, value in data["nodes"].items():
            self.comment_data[node] = {
                "CPUEff": value["total_time"]
                / value["cpus"]
                / data["total_time"]
                * 100,
                "MemEff": value["used_memory"] / value["total_memory"] * 100,
            }
            if data["gpus"]:
                self.comment_data[node]["gpus"] = {
                    gpu: {
                        "GPUEff": value["gpu_utilization"][gpu],
                        "GPUMem": round(
                            value["gpu_used_memory"][gpu]
                            / value["gpu_total_memory"][gpu]
                            * 100,
                            1,
                        ),
                    }
                    for gpu in value["gpu_utilization"]
                }
                self.comment_data[node]["GPUEff"] = average(
                    "GPUEff", self.comment_data[node]["gpus"]
                )
                self.comment_data[node]["GPUMem"] = average(
                    "GPUMem", self.comment_data[node]["gpus"]
                )

        self.cpu = average("CPUEff")
        self.mem_eff = average("MemEff")
        if data["gpus"]:
            self.gpu = average("GPUEff")
            self.gpu_mem = average("GPUMem")

    def _cache_entries(self) -> None:
        self.other_entries["State"] = self.state
        self.other_entries["TimeEff"] = self.time_eff
        self.other_entries["CPUEff"] = self.cpu if self.cpu else "---"
        self.other_entries["GPUEff"] = self.gpu if self.gpu is not None else "---"
        if self.gpu_mem is not None:
            self.other_entries["GPUMem"] = self.gpu_mem
        else:
            self.other_entries["GPUMem"] = "---"

    def name(self) -> str:
        """The name of the job.

        Returns:
            The filename (if set) or jobid
        """
        if self.filename:
            return self.filename
        return self.jobid

    def get_entry(self, key: str) -> Any:
        """Get an attribute by name.

        Args:
            key: the attribute to query

        Returns:
            The value of that attribute or "---" if not found
        """
        if key == "MemEff":
            if self.mem_eff:  # set by admin comment
                return self.mem_eff
            if self.totalmem:
                return round(self.stepmem / self.totalmem * 100, 1)
            return "---"
        return self.other_entries.get(key, "---")

    def get_node_entries(
        self, key: str, gpu: bool = False
    ) -> Generator[Any, None, None]:
        """Get an attribute by name for each node in job.

        Args:
            key: the attribute to query
            gpu: True if gpu values should also be split

        Yields:
            A generator with values for provided attribute.  If the key does
            not change in nodes/gpus it will yield an empty string
        """
        yield self.get_entry(key)
        if len(self.comment_data) > 1 or (gpu and self.gpu is not None):
            for node, data in self.comment_data.items():
                # get node-level data
                if key == "JobID":
                    yield f"  {node}"
                else:
                    to_yield = data.get(key, "")
                    to_yield = (
                        to_yield if isinstance(to_yield, str) else round(to_yield, 1)
                    )
                    yield to_yield
                if gpu and self.gpu is not None:  # has gpus to report
                    for gpu_name, gpu_data in data["gpus"].items():
                        if key == "JobID":
                            yield f"    {gpu_name}"
                        else:
                            to_yield = gpu_data.get(key, "")
                            to_yield = (
                                to_yield
                                if isinstance(to_yield, str)
                                else round(to_yield, 1)
                            )
                            yield to_yield


def _parse_slurm_timedelta(delta: str) -> int:
    """Parse one of the three formats used in TotalCPU.

    Based on which regex matches, convert into a timedelta
    and return total seconds.

    Args:
        delta: The time duration

    Returns:
        Number of seconds elapsed during the delta

    Raises:
        ValueError: if unable to parse delta
    """
    match = re.match(DDHHMMSS_RE, delta)
    if match:
        return int(
            timedelta(
                days=int(match.group("days")),
                hours=int(match.group("hours")),
                minutes=int(match.group("minutes")),
                seconds=int(match.group("seconds")),
            ).total_seconds()
        )
    match = re.match(HHMMSS_RE, delta)
    if match:
        return int(
            timedelta(
                hours=int(match.group("hours")),
                minutes=int(match.group("minutes")),
                seconds=int(match.group("seconds")),
            ).total_seconds()
        )
    match = re.match(MMSSMMM_RE, delta)
    if match:
        return int(
            timedelta(
                minutes=int(match.group("minutes")),
                seconds=int(match.group("seconds")),
                milliseconds=int(match.group("milliseconds")),
            ).total_seconds()
        )
    raise ValueError(f'Failed to parse time "{delta}"')


def parsemem(mem: str, nodes: int = 1, cpus: int = 1) -> float:
    """Parse memory representations of reqmem and maxrss.

    Args:
        mem: the memory representation
        nodes: the number of nodes in the job
        cpus: the number of cpus in the job

    Returns:
        The number of bytes for the job.
        if mem is empty, return 0.
        if mem ends with n or c, scale by the provided nodes or cpus respectively
        the multiple of memory (e.g. M or G) is always scaled if provided

    Raises:
        ValueError: if unable to parse mem
    """
    if mem == "" or mem == "0":
        return 0
    match = re.fullmatch(MEM_RE, mem)
    if not match:
        raise ValueError(f'Failed to parse memory "{mem}"')
    memory = float(match.group("memory"))

    if match.group("multiple") != "":
        memory *= multiple_map[match.group("multiple")]

    if match.group("type") != "":
        if match.group("type") == "n":
            memory *= nodes
        else:
            memory *= cpus
    return memory



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['reportseff']

package_data = \
{'': ['*']}

install_requires = \
['click>=6.7']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=4.8.2,<5.0.0']}

entry_points = \
{'console_scripts': ['reportseff = reportseff.console:main']}

setup_kwargs = {
    'name': 'reportseff',
    'version': '2.4.2',
    'description': 'Tablular seff output',
    'long_description': '[![Tests](https://github.com/troycomi/reportseff/workflows/Tests/badge.svg)](https://github.com/troycomi/reportseff/actions?workflow=Tests)\n[![codecov](https://codecov.io/gh/troycomi/reportseff/branch/main/graph/badge.svg)](https://codecov.io/gh/troycomi/reportseff)\n[![PyPI](https://img.shields.io/pypi/v/reportseff.svg)](https://pypi.org/project/reportseff/)\n\n# `reportseff`\n\n> A python script for tabular display of slurm efficiency information\n\n![Example](https://github.com/troycomi/reportseff/raw/main/imgs/example.png)\n\n## About\n\n### Motivation\n\nWhether a sys admin or cluster user, knowing how well you are estimating job\nresources can help streamline job scheduling and maximize your priority. If you\nhave ever tried to use `sacct` you probably had some trouble interpreting the\noutput.  While `seff` or `jobstats` can provide detailed summaries, they don\'t\nscale easily to array jobs or offer a way to see all the jobs from a single\nuser.  `reportseff` aims to fill this role.  Read more about the [motivation\nfor reportseff](https://github.com/troycomi/reportseff/blob/main/ABOUT.md).\n\n### Audience\n\nIf you are running more than one slurm job at a time, you should try\n`reportseff`.  Users of HPC systems can get an idea how well they estimate\nresource usage.  By tuning these values, you can get scheduled earlier and not\nbe penalized for unused allocations.  Since `reportseff` can parse job ids from\nslurm output files, it simplifies the task of identifying which jobs have\nfailed and why. Sys admins can pipe `reportseff` output to identify users with\npoor utilization or produce summaries at the end of a billing cycle.\n\n### Implementation\n\n`reportseff` is a wrapper around `sacct` that provides more complex option\nparsing, simpler options, and cleaner, colored outputs.  All querying is\nperformed in a single call to `sacct` and should have similar performance.\nMulti-node and GPU utilization is acquired from information contained in the\n`AdminComment` field, as generated by `jobstats`.\n\n## Usage\n\n### Installation\n\n`reportseff` runs on python >= 3.6.\nThe only external dependency is click (>= 6.7).\nCalling\n\n```sh\npip install --user reportseff\n# OR\npipx install reportseff\n```\n\nwill create command line bindings and install click.\n\n### Sample Usage\n\nTry `reportseff -u $USER` or just `reportseff` in a directory with some slurm\noutputs.  You may be surprised by your results!\n\n#### Single job\n\nCalling `reportseff` with a single jobid will provide equivalent information to\nseff for that job. `reportseff 24371789` and `reportseff map_fastq_24371789`\nproduce the following output:\n\n```txt\n   JobID      State          Elapsed   CPUEff   MemEff\n24371789    COMPLETED       03:08:03   71.2%    45.7%\n```\n\n#### Single array job\n\nProviding either the raw job id or the array job id will get efficiency\ninformation for a single element of the array job. `reportseff 24220929_421`\nand `reportseff 24221219` generate:\n\n```txt\n       JobID      State          Elapsed    CPUEff   MemEff\n24220929_421    COMPLETED       00:09:34    99.0%    34.6%\n```\n\n#### Array job group\n\nIf the base job id of an array is provided, all elements of the array will\nbe added to the output. `reportseff 24220929`\n\n```txt\n          JobID      State          Elapsed    CPUEff   MemEff\n     24220929_1    COMPLETED       00:10:43    99.2%    33.4%\n    24220929_11    COMPLETED       00:10:10    99.2%    37.5%\n    24220929_21    COMPLETED       00:09:25    98.8%    36.1%\n    24220929_31    COMPLETED       00:09:19    98.9%    33.3%\n    24220929_41    COMPLETED       00:09:23    98.9%    33.3%\n    24220929_51    COMPLETED       00:08:02    98.5%    36.3%\n    ...\n   24220929_951    COMPLETED       00:25:12    99.5%    33.5%\n   24220929_961    COMPLETED       00:39:26    99.7%    34.1%\n   24220929_971    COMPLETED       00:24:11    99.5%    34.2%\n   24220929_981    COMPLETED       00:24:50    99.5%    44.3%\n   24220929_991    COMPLETED       00:13:05    98.7%    33.7%\n```\n\n#### Glob expansion of slurm outputs\n\nBecause slurm output files can act as job id inputs, the following can\nget all seff information for a given job name:\n\n```txt\nslurm_out  ❯❯❯ reportseff split_ubam_24*\n              JobID      State          Elapsed   CPUEff   MemEff\nsplit_ubam_24342816    COMPLETED       23:30:32   99.9%    4.5%\nsplit_ubam_24342914    COMPLETED       22:40:51   99.9%    4.6%\nsplit_ubam_24393599    COMPLETED       23:43:36   99.4%    4.4%\nsplit_ubam_24393655    COMPLETED       21:36:58   99.3%    4.5%\nsplit_ubam_24418960     RUNNING        02:53:11    ---      ---\nsplit_ubam_24419972     RUNNING        01:26:26    ---      ---\n```\n\n#### No arguments\n\nWithout arguments, reportseff will try to find slurm output files in the\ncurrent directory. Combine with `watch` to monitor job progress:\n`watch -cn 300 reportseff --modified-sort`\n\n```txt\n                JobID           State          Elapsed   CPUEff   MemEff\n   split_ubam_24418960          RUNNING        02:56:14    ---      ---\nfastq_to_ubam_24419971          RUNNING        01:29:29    ---      ---\n   split_ubam_24419972          RUNNING        01:29:29    ---      ---\nfastq_to_ubam_24393600         COMPLETED     1-02:00:47   58.3%    41.1%\n    map_fastq_24419330          RUNNING        02:14:53    ---      ---\n    map_fastq_24419323          RUNNING        02:15:24    ---      ---\n    map_fastq_24419324          RUNNING        02:15:24    ---      ---\n    map_fastq_24419322          RUNNING        02:15:24    ---      ---\nmark_adapters_24418437         COMPLETED       01:29:23   99.8%    48.2%\nmark_adapters_24418436         COMPLETED       01:29:03   99.9%    47.4%\n```\n\n#### Filtering slurm output files\n\nOne useful application of `reportseff` is filtering a directory of slurm output\nfiles based on the state or time since running. Additionally, if only the\n`jobid` is specified as a format output, the filenames will be returned in a\npipe-friendly manner:\n\n```txt\nold_runs   ❯❯❯ reportseff --since d=4 --state Timeout\n\n                 JobID   State      Elapsed  CPUEff   MemEff\ncall_variants_31550458  TIMEOUT    20:05:17  99.5%     0.0%\ncall_variants_31550474  TIMEOUT    20:05:17  99.6%     0.0%\ncall_variants_31550500  TIMEOUT    20:05:08  99.7%     0.0%\nold_runs   ❯❯❯ reportseff --since d=4 --state Timeout --format jobid\ncall_variants_31550458\ncall_variants_31550474\ncall_variants_31550500\n```\n\nTo find all lines with `output:` in jobs which have timed out or failed\nin the last 4 days:\n\n```sh\nreportseff --since \'d=4\' --state TO,F --format jobid | xargs grep output:\n```\n\n### Arguments\n\nJobs can be passed as arguments in the following ways:\n\n- Job ID such as 1234567.  If the id is part of an array job, only the element\nfor that ID will be displayed.  If the id is the base part of an array job,\nall elements in the array will be displayed.\n- Array Job ID such as 1234567\\_89.  Will display only the element specified.\n- Slurm output file.  Format must be BASE\\_%A\\_%a.  BASE is optional as is a\n\'.out\' suffix.  Unix glob expansions can also be used to filter which jobs\nare displayed.\n- From current directory.  If no argument is supplied, `reportseff` will attempt\nto find slurm output files in the current directory as described above.\nIf a user is provided, instead `reportseff` will show recent jobs for that user.\nIf only `since` is set, all recent jobs for all users will be shown (if allowed).\n- Supplying a directory as a single argument will override the current\ndirectory to check for slurm outputs.\n\n### Options\n\n- `--color/--no-color`: Force color output or not.  By default, will force color\n  output.  With the no-color flag, click will strip color codes for everything\n  besides stdout.\n- `--modified-sort`: Instead of sorting by filename/jobid, sort by last\n  modification time of the slurm output file.\n- `--debug`: Write sacct result to stderr.\n- `--user/-u`: Ignore job arguments and instead query sacct with provided user.\n  Returns all jobs from the last week.\n- `--state/-s`: Output only jobs with states matching one of the provided options.\n  Accepts comma separated values of job codes (e.g. \'R\') or full names\n  (e.g. RUNNING).  Case insensitive.\n- `--not-state/-S`: Output only jobs with states not matching any of the provided options.\n  Accepts comma separated values of job codes (e.g. \'R\') or full names\n  (e.g. RUNNING).  Case insensitive.\n- `--format`: Provide a comma separated list of columns to produce. Prefixing the\n  argument with `+` adds the specified values to the defaults.  Values can\n  be any valid column name to sacct and the custom efficiency values: TimeEff,\n  cpuEff, MemEff.  Can also optionally set alignment (<, ^, >) and maximum width.\n  Default is center-aligned with a width of the maximum column entry.  For\n  example, `--format \'jobid%>,state%10,memeff%<5\'` produces 3 columns with:\n  - JobId aligned right, width set automatically\n  - State with width 10 (center aligned by default)\n  - MemEff aligned left, width 5\n- `--since`: Limit results to those occurring after the specified time.  Accepts\n  sacct formats and a comma separated list of key/value pairs.  To get jobs in\n  the last hour and a half, can pass `h=1,m=30`.\n- `--node/-n`: Display information for multi-node jobs; requires additional\n  sacct fields from jobstats.\n- `--node-and-gpu/-g`: Display information for multi-node jobs and GPU information;\n  requires additional sacct fields from jobstats.\n- `--parsable/-p`: Ignore formatting and output as a `|` delimited table.  Useful\n  for piping into more complex analyses.\n\n## Status, Contributions, and Support\n\n`reportseff` is actively maintained but currently feature complete.  If there\nis a function missing, please open an issue to discuss its merit!\n\nBug reports, pull requests, and any feedback are welcome! Prior to submitting\na pull request, be sure any new features have been tested and all unit tests\nare passing. In the cloned repo with\n[poetry](https://github.com/python-poetry/poetry#installation) installed:\n\n```sh\npoetry install\npoetry run pytest\npoetry run pre-commit install\nnox\n```\n\n## Troubleshooting\n\n### I can\'t install, what is pip?\n\n[pip](https://pip.pypa.io/en/stable/) is the package installer for python.  If\nyou get an error that pip isn\'t found, look for a python/anaconda/conda module.\n[pipx](https://pypa.github.io/pipx/) ensures that each application is installed\nin an isolated environment.  This resolves issues of dependency versions and\nallows applications to be run from any environment.\n\n### I get an error about broken pipes when chaining to other commands\n\nPython will report that the consumer of process output has closed the stream\n(i.e. the pipe) while still attempting to write.  Newer versions of click\nshould suppress the warning output, but it seems to not always work.  Besides\nsome extra printing on stderr, the output is not affected.\n\n### My jobs don\'t have any information about multiple nodes or GPU efficiency\n\nBecause `sacct` doesn\'t currently record this information, `reportseff`\nretrieves it from a custom field from `jobstats`, developed at Princeton\nUniversity.  If you are outside a Research Computing cluster, that information\nwill likely be absent.  Node-level reporting is only shown for jobs which use\nmultiple nodes or GPUs.  If you need a list of where jobs were run, you can add\n`--format +NodeList`.\n\n### My output is garbled with ESC[0m all over, where\'s the color?\n\nThose are ANSI color codes.  Click will usually strip these if it detects\nthe consuming process can\'t display color codes, but `reportseff` defaults\nto always display them.  If you don\'t care for color, use the `--no-color`\noption.  For less, you can set\n```\nexport LESS="-R"\n```\nin your `.bashrc`, or just type `-R` in an active less process.  Some versions\nof `watch` require the `-c` option to display color, others can\'t display\ncolors properly.  If you search for `ansi color <tool>` you should get some\nsolutions.\n\n## Acknowledgments\n\nThe code for calling sacct and parsing the returning information was taken\nfrom [Slurmee](https://github.com/PrincetonUniversity/slurmee).\n\nStyle and tooling from [hypermodern python](https://cjolowicz.github.io/posts/hypermodern-python-01-setup/)\n\nCode review provided from a [repo-review](https://researchcomputing.princeton.edu/services/repo-review-consultations)\nwhich vastly improved this readme.\n',
    'author': 'Troy Comi',
    'author_email': 'troycomi@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/troycomi/reportseff',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<4.0',
}


setup(**setup_kwargs)



"""Command line parameter collection."""


class ReportseffParameters:
    """Collection of parameters from the command line.

    Basically a dataclass but I want to be python 3.6 compatible without the
    pip package
    """

    color: bool
    debug: bool = False
    format_str: str = ""
    jobs: tuple
    modified_sort: bool = False
    node: bool = False
    node_and_gpu: bool = False
    not_state: str = ""
    parsable: bool = False
    since: str = ""
    until: str = ""
    state: str = ""
    user: str = ""
    partition: str = ""

    def __init__(
        self,
        jobs: tuple,
        color: bool = True,
        debug: bool = False,
        format_str: str = "",
        modified_sort: bool = False,
        node: bool = False,
        node_and_gpu: bool = False,
        not_state: str = "",
        parsable: bool = False,
        since: str = "",
        until: str = "",
        state: str = "",
        user: str = "",
        partition: str = "",
    ) -> None:
        """Create a new parameter object.

        Args:
            jobs: job ids, files, or directory to query
            color: if the output should be colored
            debug: if debug information should be printed
            format_str: format string specification
            modified_sort: if output should be sorted by modified time
            node: if node information should be displayed
            node_and_gpu: if node and gpu information should be displayed
            not_state: states to remove
            parsable: if output should be parsable
            since: display jobs since a specific time
            until: display jobs until a specific time
            state: stats to display
            user: the user to report for
            partition: the partition to report for
        """
        self.color = color
        self.debug = debug
        self.format_str = format_str
        self.jobs = jobs
        self.modified_sort = modified_sort
        self.node = node
        self.node_and_gpu = node_and_gpu
        self.not_state = not_state
        self.parsable = parsable
        self.since = since
        self.until = until
        self.state = state
        self.user = user
        self.partition = partition

        if self.format_str.startswith("+"):
            self.format_str = (
                "JobID%>,State,Elapsed%>,TimeEff,CPUEff,MemEff," + self.format_str[1:]
            )
